/*
* Safe Eyes is a utility to remind you to take break frequently
* to protect your eyes from eye strain.

* Copyright (C) 2016  Gobinath

* This program is free software: you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 of the License, or
* (at your option) any later version.

* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.

* You should have received a copy of the GNU General Public License
* along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

.window_main {
    background: black;
    opacity: 0.9;
    border-color: transparent;
}

.btn_skip {
    color: white;
    font-size: 10pt;
    border-radius: 25px;
    padding-top: 10px;
    padding-bottom: 10px;
    padding-left: 25px;
    padding-right: 25px;
    border-color: white;
    background: transparent;
    border-width: 2px;
    border-image: none;
}

.btn_skip:hover {
    background: white;
    color: black;
}

.btn_postpone {
    color: white;
    font-size: 10pt;
    border-radius: 25px;
    padding-top: 10px;
    padding-bottom: 10px;
    padding-left: 25px;
    padding-right: 25px;
    border-color: white;
    background: transparent;
    border-width: 2px;
    border-image: none;
}

.btn_postpone:hover {
    background: white;
    color: black;
}

.lbl_message {
    font-size: 22pt;
    color: white;
    font-weight: bold;
}

.lbl_count {
    font-size: 12pt;
    color: white;
}

.lbl_widget {
    font-size: 9pt;
    color: white;
}

.btn_circle {
    border-radius: 25px;
}

.lbl_plugin_name {
    font-weight: bold;
}

.lbl_plugin_description {
    color: #9B9B9B;
}

.btn_plugin_extra_link {
    padding-left: 0px;
}

.info_bar_long_break {
    opacity: 0.9;
    border-radius: 5px;
}

.warn_bar_rpc_server {
    opacity: 0.9;
    border-radius: 5px;
}

.toolbar {
    background: black;
    opacity: 0.9;
    border-color: transparent;
}

.btn_menu {
    border-width: 0px;
    border-radius: 0px;
    border-image: None;
    background: white;
    border-color: transparent;
}

.btn_menu:hover {
    border-width: 0px;
    border-radius: 0px;
    border-image: None;
    background: whitesmoke;
    border-color: transparent;
}
